var run=require('child_process').spawn;

module.exports = function(RED) {

	function FunctionNode(n) {
		RED.nodes.createNode(this,n);
		var node = this;
		this.name = n.name;
		this.func = n.func;
		this.rqs=n.rqs;
		this.persistent=n.persistent;
		this.threads=n.threads;
		var functionText = "var results = null;"+
				   "results = (function(msg){ "+
					  "var __msgid__ = msg._msgid;"+
					  "var node = {"+
						 "log:__node__.log,"+
						 "error:__node__.error,"+
						 "warn:__node__.warn,"+
						 "status:__node__.status,"+
						 "send:function(msgs){ __node__.send(__msgid__,msgs);}"+
					  "};\n"+
					  this.func+"\n"+
				   "})(msg);";
		this.topic = n.topic;
		this.workers=[];
		this.cleanup=false;
		this.taskQueue=[];
		function resumeSomeDelayedMessages(){
			if (node.persistent){
				node.workers.filter(a=>!a.busy).forEach(w=>{
					if (node.taskQueue.length>0){
						w.busy=true;
						w.send({input:node.taskQueue.shift()});
					}
				});
			}
			else {
				while (node.taskQueue.length>0 && node.workers.length<node.threads){
					let nW=startWorker();
					nW.busy=true;
					nW.send({input:node.taskQueue.shift()})
				}
			}
		}
		function startWorker(){
			let nW=run(process.argv[0],[module.filename.replace("mpf","mpf_w"),node.name],{stdio:['ignore',require('fs').openSync('./mpf.log','a+'),require('fs').openSync('./mpf.err','a+'),'ipc']});//,'pipe','pipe'
			nW.busy=false;
			node.workers.push(nW);
			nW.on('message',(msg)=>{
				if (!!msg.log) node.log.apply(node, msg.log);
				if (!!msg.warn) node.warn.apply(node, msg.warn);
				if (!!msg.error) node.error.apply(node, msg.error);
				if (!!msg.status) node.status.apply(node, msg.status);
				if (!!msg.output) {
					if (msg.output instanceof Array) msg.output.forEach(a=>{if (!!a) a.doneByPid=nW.pid});
					else if (!!msg) msg.output.doneByPid=nW.pid;
					node.send(msg.output);
				}
				if (!!msg.idle){
					nW.busy=false;
					setTimeout(resumeSomeDelayedMessages,100);
				}
			});
			nW.on('close',(code)=>{
				node[!!node.cleanup||!node.persistent?"log":"error"]("worker exited with code "+code);
				node.workers=node.workers.filter(a=>a.pid!=nW.pid);
				if (!node.cleanup)
					setTimeout(startWorkers,100);
			});
			nW.send({
				setup:{
					functionText:functionText,
					rqs:node.rqs,
					persistent:node.persistent
				}
			});
			return nW;
		}
		function startWorkers(){
			if (node.persistent) 
				while (node.workers.length<node.threads)
					startWorker();
		}
		try {
			startWorkers();
			this.on("input", function(msg) {
				if (node.persistent){
					let freeW=node.workers.filter(a=>!a.busy);
					if (freeW.length>0){
						freeW[0].busy=true;
						freeW[0].send({input:msg})
					}
					else {
						node.taskQueue.push(msg);
					}
				}
				else {
					if (node.workers.length<node.threads){
						let nW=startWorker();
						nW.busy=true;
						nW.send({input:msg});
					}
					else {
						node.taskQueue.push(msg);
					}
				}
			});
			this.on("close", function() {
				node.cleanup=true;
				node.workers.forEach(w=>w.kill());
				node.workers=[];
				node.status({});
			})
		} catch(err) {
			this.error(err);
		}
	}
	RED.nodes.registerType("mp-function",FunctionNode);
}
