var util = require("util");
var vm = require("vm");

function sendResults(_msgid,msgs) {
	if (msgs == null) {
		return;
	} else if (!util.isArray(msgs)) {
		msgs = [msgs];
	}
	var msgCount = 0;
	for (var m=0;m<msgs.length;m++) {
		if (msgs[m]) {
			if (util.isArray(msgs[m])) {
				for (var n=0; n < msgs[m].length; n++) {
					if (msgs[m][n] !== null && msgs[m][n] !== undefined) {
						msgs[m][n]._msgid = _msgid;
						msgCount++;
					}
				}
			} else {
				if (msgs[m] !== null && msgs[m] !== undefined) {
					msgs[m]._msgid = _msgid;
					msgCount++;
				}
			}
		}
	}
	if (msgCount>0) {
		process.send({output:msgs});
	}
}

var outstandingTimers = [];
var outstandingIntervals = [];
var sandbox = {
	console:console,
	util:util,
	Buffer:Buffer,
	__node__: {
		log: function() {
			process.send({log:[].slice.apply(arguments,[])});
		},
		error: function() {
			process.send({error:[].slice.apply(arguments,[])});
		},
		warn: function() {
			process.send({warn:[].slice.apply(arguments,[])});
		},
		send: function(id, msgs) {
			sendResults(id, msgs);
		},
		status: function() {
			process.send({status:[].slice.apply(arguments,[])});
		}
	},
	context: {
		global:{
		}
	},
	setTimeout: function () {
		var func = arguments[0];
		var timerId;
		arguments[0] = function() {
			sandbox.clearTimeout(timerId);
			try {
				func.apply(this,arguments);
			} catch(err) {
				node.error(err,{});
			}
		};
		timerId = setTimeout.apply(this,arguments);
		outstandingTimers.push(timerId);
		return timerId;
	},
	clearTimeout: function(id) {
		clearTimeout(id);
		var index = outstandingTimers.indexOf(id);
		if (index > -1) {
			outstandingTimers.splice(index,1);
		}
	},
	setInterval: function() {
		var func = arguments[0];
		var timerId;
		arguments[0] = function() {
			try {
				func.apply(this,arguments);
			} catch(err) {
				node.error(err,{});
			}
		};
		timerId = setInterval.apply(this,arguments);
		outstandingIntervals.push(timerId);
		return timerId;
	},
	clearInterval: function(id) {
		clearInterval(id);
		var index = outstandingIntervals.indexOf(id);
		if (index > -1) {
			outstandingIntervals.splice(index,1);
		}
	}
};
var script,persistent=false;
var setUp=false;


function processInput(msg){
	if (!setUp) setTimeout(processInput(msg),100);
	else {
		try {
			var context = vm.createContext(sandbox);
			var start = process.hrtime();
			context.msg = msg.input;
			script.runInContext(context);
			sendResults(msg.input._msgid,context.results);

			var duration = process.hrtime(start);
			var converted = Math.floor((duration[0] * 1e9 + duration[1])/10000)/100;
			if (process.env.NODE_RED_FUNCTION_TIME) {
				process.send({status:{fill:"yellow",shape:"dot",text:""+converted}});
			}
		} catch(err) {

			var line = 0;
			var errorMessage;
			var stack = err.stack.split(/\r?\n/);
			if (stack.length > 0) {
				while (line < stack.length && stack[line].indexOf("ReferenceError") !== 0) {
					line++;
				}

				if (line < stack.length) {
					errorMessage = stack[line];
					var m = /:(\d+):(\d+)$/.exec(stack[line+1]);
					if (m) {
						var lineno = Number(m[1])-1;
						var cha = m[2];
						errorMessage += " (line "+lineno+", col "+cha+")";
					}
				}
			}
			if (!errorMessage) {
				errorMessage = err.toString();
			}
			process.send({error:[errorMessage, msg.input]});
		}
		process.send({idle:true});
		if (!persistent) process.exit(0);
	}
}

process.on('message',function(msg){
	if (!!msg.setup){
		if (!!msg.setup.rqs) msg.setup.rqs.split(",").map(a=>a.split("<-")).forEach(a=>{
			sandbox.context.global[a[0]]=require(a[1]);
		}) 
		script = vm.createScript(msg.setup.functionText);
		persistent=msg.setup.persistent;
		setUp=true;
	}
	if (!!msg.input){
		processInput(msg);
	}
});

process.on('disconnect',()=>process.exit(0))